/****************************************************************************
**
** Copyright (C) 2008 Nokia Corporation and/or its subsidiary(-ies).
** Contact: Qt Software Information (qt-info@nokia.com)
**
** This file is part of the QtCore module of the Qt Toolkit.
**
** No Commercial Usage
** This file contains pre-release code and may only be used for
** evaluation and testing purposes.  It may not be used for commercial
** development.  You may use this file in accordance with the terms and
** conditions contained in the either Technology Preview License
** Agreement or the Beta Release License Agreement.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License versions 2.0 or 3.0 as published by the Free
** Software Foundation and appearing in the file LICENSE.GPL included in
** the packaging of this file.  Please review the following information
** to ensure GNU General Public Licensing requirements will be met:
** http://www.fsf.org/licensing/licenses/info/GPLv2.html and
** http://www.gnu.org/copyleft/gpl.html.  In addition, as a special
** exception, Nokia gives you certain additional rights. These rights
** are described in the Nokia Qt GPL Exception version 1.3, included in
** the file GPL_EXCEPTION.txt in this package.
**
** Qt for Windows(R) Licensees
** As a special exception, Nokia, as the sole copyright holder for Qt
** Designer, grants users of the Qt/Eclipse Integration plug-in the
** right for the Qt/Eclipse Integration to link to functionality
** provided by Qt Designer and its related libraries.
**
** If you are unsure which license is appropriate for your use, please
** contact the sales department at qt-sales@nokia.com.
**
****************************************************************************/

#include "qsharedpointer.h"

// to be sure we aren't causing a namespace clash:
#include "qshareddata.h"

/*!
    \class QSharedPointer
    \brief The QSharedPointer class holds a strong reference to a shared pointer
    \since 4.5

    \reentrant
    \ingroup misc

    The QSharedPointer is an automatic, shared pointer in C++. It
    behaves exactly like a normal pointer for normal purposes,
    including respect for constness.

    QSharedPointer will delete the pointer it is holding when it goes
    out of scope, provided no other QSharedPointer objects are
    referencing it.

    A QSharedPointer object can be created from a normal pointer,
    another QSharedPointer object or by promoting a
    QWeakPointer object to a strong reference.

    \section Thread-safety

    QSharedPointer and QWeakPointer are thread-safe and operate
    atomically on the pointer value. Different threads can also access
    the same QSharedPointer or QWeakPointer object at the same time
    without need for locking mechanisms.

    It should be noted that, while the pointer value can be accessed
    in this manner, QSharedPointer and QWeakPointer provide no
    guarantee about the object being pointed to. Thread-safety and
    reentrancy rules for that object still apply.

    \section Other pointer classes

    Qt also provides two other pointer wrapper classes: QPointer and
    QSharedDataPointer. They are incompatible with one another, since
    each has its very different use case.

    QSharedPointer holds a shared pointer by means of an external
    reference count (i.e., a reference counter placed outside the
    object). Like its name indicates, the pointer value is shared
    among all instances of QSharedPointer and QWeakPointer. The
    contents of the object pointed to by the pointer should not
    considered shared, however: there is only one object. For that
    reason, QSharedPointer does not provide a way to detach or make
    copies of the pointed object.

    QSharedDataPointer, on the other hand, holds a pointer to shared
    data (i.e., a class derived from QSharedData). It does so by means
    of an internal reference count, placed in the QSharedData base
    class. This class can, therefore, detach based on the type of
    access made to the data being guarded: if it's a non-const access,
    it creates a copy atomically for the operation to complete.

    QExplicitlySharedDataPointer behaves like QSharedDataPointer,
    except that it only detaches if
    QExplicitlySharedDataPointer::detach() is explicitly called.

    Finally, QPointer holds a pointer to a QObject-derived object, but
    it does so weakly. QPointer is similar, in that behaviour, to
    QWeakPointer: it does not allow you to prevent the object from
    being destroyed. All you can do is query whether it has been
    destroyed or not.

    \sa QSharedDataPointer, QWeakPointer
*/

/*!
    \class QWeakPointer
    \brief The QWeakPointer class holds a weak reference to a shared pointer
    \since 4.5
    \reentrant
    \ingroup misc

    The QWeakPointer is an automatic weak reference to a
    pointer in C++. It cannot be used to dereference the pointer
    directly, but it can be used to verify if the pointer has been
    deleted or not in another context.

    QWeakPointer objects can only be created by assignment
    from a QSharedPointer.

    To access the pointer that QWeakPointer is tracking, you
    must first create a QSharedPointer object and verify if the pointer
    is null or not.

    \sa QSharedPointer
*/

/*!
    \fn QSharedPointer::QSharedPointer()

    Creates a QSharedPointer that points to null (0).
*/

/*!
    \fn QSharedPointer::~QSharedPointer()

    Destroys this QSharedPointer object. If it is the last reference to
    the pointer stored, this will delete the pointer as well.
*/

/*!
    \fn QSharedPointer::QSharedPointer(T *ptr)

    Creates a QSharedPointer that points to \a ptr. The pointer \a ptr
    becomes managed by this QSharedPointer and must not be passed to
    another QSharedPointer object or deleted outside this object.
*/

/*!
    \fn QSharedPointer::QSharedPointer(const QSharedPointer<T> &other)

    Creates a QSharedPointer object that shares \a other's pointer.

    If \tt T is a derived type of the template parameter of this class,
    QSharedPointer will perform an automatic cast. Otherwise, you will
    get a compiler error.

    \warning Automatic downcasting is not available on MSVC 6. Use
    qSharedPointerCast() for portability to MSVC 6.
*/

/*!
    \fn QSharedPointer::QSharedPointer(const QWeakPointer<T> &other)

    Creates a QSharedPointer by promoting the weak reference \a other
    to strong reference and sharing its pointer.

    If \tt T is a derived type of the template parameter of this
    class, QSharedPointer will perform an automatic cast. Otherwise,
    you will get a compiler error.

    \warning Automatic downcasting is not available on MSVC 6. Use
    qSharedPointerCast() for portability to MSVC 6.
*/

/*!
    \fn QSharedPointer &QSharedPointer::operator=(T *ptr)

    Makes this object point to \a ptr. The current pointer reference
    is discarded and, if it was the last, the pointer will be deleted.
*/

/*!
    \fn QSharedPointer &QSharedPointer::operator=(const QSharedPointer<T> &other)

    Makes this object share \a other's pointer. The current pointer
    reference is discarded and, if it was the last, the pointer will
    be deleted.

    If \tt T is a derived type of the template parameter of this
    class, QSharedPointer will perform an automatic cast. Otherwise,
    you will get a compiler error.

    \warning Automatic downcasting is not available on MSVC 6. Use
    qSharedPointerCast() for portability to MSVC 6.
*/

/*!
    \fn QSharedPointer &QSharedPointer::operator=(const QWeakPointer<T> &other)

    Promotes \a other to a strong reference and makes this object
    share a reference to the pointer referenced by it. The current pointer
    reference is discarded and, if it was the last, the pointer will
    be deleted.

    If \tt T is a derived type of the template parameter of this
    class, QSharedPointer will perform an automatic cast. Otherwise,
    you will get a compiler error.

    \warning Automatic downcasting is not available on MSVC 6. Use
    qSharedPointerCast() for portability to MSVC 6.
*/

/*!
    \fn bool QSharedPointer::operator==(const T *ptr) const

    Returns true if the pointer referenced by this object and \a ptr
    are the same.
*/

/*!
    \fn bool QSharedPointer::operator!=(const T *ptr) const

    Returns true if the pointer referenced by this object and \a ptr
    are different.
*/

/*!
    \fn bool QSharedPointer::operator==(const QSharedPointer<T> &other) const

    Returns true if the pointer referenced by this object and is the
    same pointer as that referenced by \a other.

    If \a other's template parameter is different from this object's,
    QSharedPointer will attempt to perform an automatic \tt static_cast
    to ensure that the pointers being compared are equal. If \a other's
    template parameter is not a base or a derived type from this
    object's, you will get a compiler error.

    \warning Automatic casting is not available on MSVC 6. If you
    intend to deploy code on MSVC 6, you need to first use
    qSharedPointerCast() to cast to the same type first.
*/

/*!
    \fn bool QSharedPointer::operator!=(const QSharedPointer<T> &other) const

    Returns true if the pointer referenced by this object is not the
    same pointer as that referenced by \a other.

    If \a other's template parameter is different from this object's,
    QSharedPointer will attempt to perform an automatic \tt static_cast
    to ensure that the pointers being compared are not equal. If \a other's
    template parameter is not a base or a derived type from this
    object's, you will get a compiler error.

    \warning Automatic casting is not available on MSVC 6. If you
    intend to deploy code on MSVC 6, you need to first use
    qSharedPointerCast() to cast to the same type first.
*/

/*!
    \fn bool QSharedPointer::operator==(const QWeakPointer<T> &other) const

    Returns true if the pointer referenced by this object and is the
    same pointer as that referenced by \a other.

    If \a other's template parameter is different from this object's,
    QSharedPointer will attempt to perform an automatic \tt static_cast
    to ensure that the pointers being compared are equal. If \a other's
    template parameter is not a base or a derived type from this
    object's, you will get a compiler error.

    \warning Automatic casting is not available on MSVC 6. If you
    intend to deploy code on MSVC 6, you need to first use
    qWeakPointerCast() to cast to the same type first.
*/

/*!
    \fn bool QSharedPointer::operator!=(const QWeakPointer<T> &other) const

    Returns true if the pointer referenced by this object is not the
    same pointer as that referenced by \a other.

    If \a other's template parameter is different from this object's,
    QSharedPointer will attempt to perform an automatic \tt static_cast
    to ensure that the pointers being compared are not equal. If \a other's
    template parameter is not a base or a derived type from this
    object's, you will get a compiler error.

    \warning Automatic casting is not available on MSVC 6. If you
    intend to deploy code on MSVC 6, you need to first use
    qWeakPointerCast() to cast to the same type first.
*/

/*!
    \fn T *QSharedPointer::data() const

    Returns the value of the pointer referenced by this object.

    Note: do not delete the pointer returned by this function or pass
    it to another function that could delete it, including creating
    QSharedPointer or QWeakPointer objects.
*/

/*!
    \fn T &QSharedPointer::operator *() const

    Provides access to the shared pointer's members.

    \sa isNull()
*/

/*!
    \fn T *QSharedPointer::operator ->() const

    Provides access to the shared pointer's members.

    \sa isNull()
*/

/*!
    \fn bool QSharedPointer::isNull() const

    Returns true if this object is holding a reference to a null
    pointer.
*/

/*!
    \fn QSharedPointer::operator bool() const

    Returns true if this object is not null. This function is suitable
    for use in \tt if-constructs, like:

    \code
        if (sharedptr) { ... }
    \endcode

    \sa isNull()
*/

/*!
    \fn bool QSharedPointer::operator !() const

    Returns true if this object is null. This function is suitable
    for use in \tt if-constructs, like:

    \code
        if (!sharedptr) { ... }
    \endcode

    \sa isNull()
*/

/*!
    \fn QSharedPointer<X> QSharedPointer::staticCast() const

    Performs a static cast from this pointer's type to \tt X and returns
    a QSharedPointer that shares the reference. This function can be
    used for up- and for down-casting, but is more useful for
    up-casting.

    \warning This function is not available in MSVC 6. Use
    qSharedPointerCast() for code intended to build on that compiler.
*/

/*!
    \fn QSharedPointer<X> QSharedPointer::dynamicCast() const

    Performs a dynamic cast from this pointer's type to \tt X and
    returns a QSharedPointer that shares the reference. If this
    function is used to up-cast, then QSharedPointer will perform a \tt
    dynamic_cast, which means that if the object being pointed by this
    QSharedPointer is not of type \tt X, the returned object will be
    null.

    \warning This function is not available in MSVC 6. Use
    qSharedPointerDynamicCast() for code intended to build on that compiler.

    \sa isNull()
*/

/*!
    \fn QWeakPointer<T> QSharedPointer::toWeakRef() const

    Returns a weak reference object that shares the pointer referenced
    by this object.
*/

/*!
    \fn QWeakPointer::QWeakPointer()

    Creates a QWeakPointer that points to nothing.
*/

/*!
    \fn QWeakPointer::~QWeakPointer()

    Destroys this QWeakPointer object. The pointer referenced
    by this object will not be deleted.
*/

/*!
    \fn QWeakPointer::QWeakPointer(const QWeakPointer<T> &other)

    Creates a QWeakPointer that holds a weak reference to the
    pointer referenced by \a other.

    If \tt T is a derived type of the template parameter of this
    class, QWeakPointer will perform an automatic cast. Otherwise,
    you will get a compiler error.

    \warning Automatic downcasting is not available on MSVC 6. Use
    qWeakPointerCast() for portability to MSVC 6.
*/

/*!
    \fn QWeakPointer::QWeakPointer(const QSharedPointer<T> &other)

    Creates a QWeakPointer that holds a weak reference to the
    pointer referenced by \a other.

    If \tt T is a derived type of the template parameter of this
    class, QWeakPointer will perform an automatic cast. Otherwise,
    you will get a compiler error.

    \warning Automatic downcasting is not available on MSVC 6. Use
    qWeakPointerCast() for portability to MSVC 6.
*/

/*!
    \fn QWeakPointer &QWeakPointer::operator=(const QWeakPointer<T> &other)

    Makes this object share \a other's pointer. The current pointer
    reference is discarded but is not deleted.

    If \tt T is a derived type of the template parameter of this
    class, QWeakPointer will perform an automatic cast. Otherwise,
    you will get a compiler error.

    \warning Automatic downcasting is not available on MSVC 6. Use
    qWeakPointerCast() for portability to MSVC 6.
*/

/*!
    \fn QWeakPointer &QWeakPointer::operator=(const QSharedPointer<T> &other)

    Makes this object share \a other's pointer. The current pointer
    reference is discarded but is not deleted.

    If \tt T is a derived type of the template parameter of this
    class, QWeakPointer will perform an automatic cast. Otherwise,
    you will get a compiler error.

    \warning Automatic downcasting is not available on MSVC 6. Use
    qWeakPointerCast() for portability to MSVC 6.
*/

/*!
    \fn bool QWeakPointer::operator==(const QWeakPointer<T> &other) const

    Returns true if the pointer referenced by this object and is the
    same pointer as that referenced by \a other.

    If \a other's template parameter is different from this object's,
    QWeakPointer will attempt to perform an automatic \tt static_cast
    to ensure that the pointers being compared are equal. If \a other's
    template parameter is not a base or a derived type from this
    object's, you will get a compiler error.

    \warning Automatic casting is not available on MSVC 6. If you
    intend to deploy code on MSVC 6, you need to first use
    qWeakPointerCast() to cast to the same type first.
*/

/*!
    \fn bool QWeakPointer::operator!=(const QWeakPointer<T> &other) const

    Returns true if the pointer referenced by this object is not the
    same pointer as that referenced by \a other.

    If \a other's template parameter is different from this object's,
    QWeakPointer will attempt to perform an automatic \tt static_cast
    to ensure that the pointers being compared are not equal. If \a other's
    template parameter is not a base or a derived type from this
    object's, you will get a compiler error.

    \warning Automatic casting is not available on MSVC 6. If you
    intend to deploy code on MSVC 6, you need to first use
    qWeakPointerCast() to cast to the same type first.
*/

/*!
    \fn bool QWeakPointer::operator==(const QSharedPointer<T> &other) const

    Returns true if the pointer referenced by this object and is the
    same pointer as that referenced by \a other.

    If \a other's template parameter is different from this object's,
    QWeakPointer will attempt to perform an automatic \tt static_cast
    to ensure that the pointers being compared are equal. If \a other's
    template parameter is not a base or a derived type from this
    object's, you will get a compiler error.

    \warning Automatic casting is not available on MSVC 6. If you
    intend to deploy code on MSVC 6, you need to first use
    qWeakPointerCast() to cast to the same type first.
*/

/*!
    \fn bool QWeakPointer::operator!=(const QSharedPointer<T> &other) const

    Returns true if the pointer referenced by this object is not the
    same pointer as that referenced by \a other.

    If \a other's template parameter is different from this object's,
    QWeakPointer will attempt to perform an automatic \tt static_cast
    to ensure that the pointers being compared are not equal. If \a other's
    template parameter is not a base or a derived type from this
    object's, you will get a compiler error.

    \warning Automatic casting is not available on MSVC 6. If you
    intend to deploy code on MSVC 6, you need to first use
    qWeakPointerCast() to cast to the same type first.
*/

/*!
    \fn bool QWeakPointer::isNull() const

    Returns true if this object is holding a reference to a null
    pointer.

    Note that, due to the nature of weak references, the pointer that
    QWeakPointer references can become null at any moment, so
    the value returned from this function can change from false to
    true from one call to the next.
*/

/*!
    \fn QWeakPointer::operator bool() const

    Returns true if this object is not null. This function is suitable
    for use in \tt if-constructs, like:

    \code
        if (weakref) { ... }
    \endcode

    Note that, due to the nature of weak references, the pointer that
    QWeakPointer references can become null at any moment, so
    the value returned from this function can change from true to
    false from one call to the next.

    \sa isNull()
*/

/*!
    \fn bool QWeakPointer::operator !() const

    Returns true if this object is null. This function is suitable
    for use in \tt if-constructs, like:

    \code
        if (!weakref) { ... }
    \endcode

    Note that, due to the nature of weak references, the pointer that
    QWeakPointer references can become null at any moment, so
    the value returned from this function can change from false to
    true from one call to the next.

    \sa isNull()
*/

/*!
    \fn QWeakPointer<X> QWeakPointer::staticCast() const

    Performs a static cast from this pointer's type to \tt X and returns
    a QWeakPointer that shares the reference. This function can be
    used for up- and for down-casting, but is more useful for
    up-casting.

    Note that dynamic casting is not available for weak references
    since the value of the pointer is not stable (it can become null
    at any moment).

    \warning This function is not available in MSVC 6. Use
    qWeakPointerCast() for code intended to build on that compiler.
*/

/*!
    \fn QSharedPointer<T> QWeakPointer::toStrongRef() const

    Promotes this weak reference to a strong one and returns a
    QSharedPointer object holding that reference.
*/

/*!
    \fn QSharedPointer<X> qSharedPointerCast(const QSharedPointer<T> &src)
    \relatesto QSharedPointer

*/

/*!
    \fn QSharedPointer<X> qSharedPointerCast(const QWeakPointer<T> &src)
    \relatesto QSharedPointer

*/

/*!
    \fn QSharedPointer<X> qSharedPointerDynamicCast(const QSharedPointer<T> &src)
    \relatesto QSharedPointer

*/

/*!
    \fn QWeakPointer<X> qWeakPointerCast(const QWeakPointer<T> &src)
    \relatesto QWeakPointer

*/

/*!
    \fn QWeakPointer<X> qWeakPointerCast(const QSharedPointer<T> &src)
    \relatesto QWeakPointer

*/

#include <qset.h>
#include <qmutex.h>

#if !defined(QT_NO_PARTIAL_TEMPLATE_SPECIALIZATION) && !defined(QT_NO_MEMBER_TEMPLATES)

namespace {
    QT_USE_NAMESPACE
    class KnownPointers
    {
    public:
        QMutex mutex;
        QSet<void *> values;
    };
}

Q_GLOBAL_STATIC(KnownPointers, knownPointers)

QT_BEGIN_NAMESPACE

/*!
    \internal
*/
void QtSharedPointer::internalSafetyCheckAdd(const volatile void *ptr)
{
    QMutexLocker lock(&knownPointers()->mutex);
    void *actual = const_cast<void*>(ptr);
    if (knownPointers()->values.contains(actual))
        qFatal("QSharedPointerData: internal self-check failed: pointer %p was already tracked "
               "by another QSharedPointerData object", actual);
    knownPointers()->values.insert(actual);
}

/*!
    \internal
*/
void QtSharedPointer::internalSafetyCheckRemove(const volatile void *ptr)
{
    QMutexLocker lock(&knownPointers()->mutex);
    void *actual = const_cast<void*>(ptr);
    knownPointers()->values.remove(actual);
}

QT_END_NAMESPACE

#endif
